from io import BufferedReader
from pathlib import Path
from typing import Any, Dict, List, Optional

from pydantic import UUID4, BaseModel, Field, SecretStr, field_validator

from promptquality.constants.integrations import IntegrationName
from promptquality.constants.run import RunDefaults
from promptquality.types.settings import Settings
from promptquality.utils.dataset import DatasetType, dataset_to_path
from promptquality.utils.name import random_name


class RandomName(BaseModel):
    name: str = Field(default_factory=random_name)

    @field_validator("name", mode="before")
    def set_name(cls, value: Optional[str]) -> str:
        if value is None:
            return random_name()
        return value


class CreateProjectRequest(RandomName):
    type: str = RunDefaults.project_type


class CreateProjectResponse(CreateProjectRequest):
    id: UUID4


class BaseTemplateVersionRequest(BaseModel):
    template: str
    version: Optional[int] = None


class CreateTemplateRequest(RandomName, BaseTemplateVersionRequest):
    project_id: UUID4


class CreateTemplateVersionRequest(BaseTemplateVersionRequest):
    template_id: UUID4
    project_id: UUID4


class BaseTemplateVersionResponse(BaseTemplateVersionRequest):
    id: UUID4


class CreateTemplateVersionResponse(BaseTemplateVersionResponse):
    version: int


class BaseTemplateResponse(RandomName):
    id: UUID4
    template: str
    selected_version: CreateTemplateVersionResponse
    selected_version_id: UUID4
    all_versions: List[CreateTemplateVersionResponse] = Field(default_factory=list)


class BaseDatasetRequest(BaseModel):
    file_path: Path = Field(exclude=True)

    @field_validator("file_path", mode="before")
    def dataset_to_path(cls, value: DatasetType) -> Path:
        return dataset_to_path(value)

    @property
    def files(self) -> Dict[str, BufferedReader]:
        return dict(file=self.file_path.open("rb"))


class UploadDatasetRequest(BaseDatasetRequest):
    project_id: UUID4
    file_path: Path
    prompt_template_version_id: UUID4

    @property
    def data(self) -> Dict[str, str]:
        return dict(prompt_template_version_id=str(self.prompt_template_version_id))


class UploadDatasetResponse(BaseModel):
    id: UUID4 = Field(alias="dataset_id")


class EstimateCostRequest(BaseDatasetRequest, Settings):
    project_id: UUID4 = Field(exclude=True)
    template: str = Field(serialization_alias="prompt_template")

    @property
    def params(self) -> Dict[str, str]:
        return self.model_dump(
            mode="json", by_alias=True, exclude_unset=True, exclude_none=True
        )


class EstimatedCostResponse(BaseModel):
    total_cost: float
    query_cost: float
    log_probs_cost: float


class CreateRunRequest(RandomName):
    project_id: UUID4
    task_type: int = RunDefaults.task_type


class CreateRunResponse(CreateRunRequest):
    id: UUID4


class CreateJobRequest(BaseModel):
    project_id: UUID4
    run_id: UUID4
    prompt_dataset_id: UUID4
    prompt_template_version_id: UUID4
    prompt_settings: Optional[Settings] = None
    job_name: str = RunDefaults.job_name
    task_type: int = RunDefaults.task_type


class CreateJobResponse(CreateJobRequest):
    id: UUID4 = Field(alias="job_id")


class GetMetricsRequest(BaseModel):
    project_id: UUID4
    run_id: UUID4


class PromptMetrics(BaseModel):
    total_responses: Optional[int] = None
    average_hallucination: Optional[float] = None
    average_bleu: Optional[float] = None
    average_rouge: Optional[float] = None
    average_cost: Optional[float] = None
    total_cost: Optional[float] = None


class GetJobStatusResponse(BaseModel):
    id: UUID4
    project_id: UUID4
    run_id: UUID4
    status: str
    error_message: Optional[str]
    progress_message: Optional[str]
    steps_completed: int = 0
    steps_total: int = 0
    progress_percent: float = 0.0


class CreateIntegrationRequest(BaseModel):
    api_key: SecretStr
    name: IntegrationName = Field(default=IntegrationName.openai)
    organization_id: Optional[str] = None

    @property
    def body(self) -> Dict[str, Any]:
        extra = (
            dict(organization_id=self.organization_id)
            if self.organization_id
            else dict()
        )
        return dict(token=self.api_key.get_secret_value(), extra=extra)


class SelectTemplateVersionRequest(BaseModel):
    project_id: UUID4
    template_id: UUID4
    version: int



from typing import Optional

from pydantic import SecretStr

from promptquality.set_config import set_config
from promptquality.types.config import Config
from promptquality.types.run import CreateIntegrationRequest


def add_openai_integration(
    api_key: str, organization_id: Optional[str] = None, config: Optional[Config] = None
) -> None:
    config = config or set_config()
    config.api_client.put_integration(
        CreateIntegrationRequest(
            api_key=SecretStr(api_key), organization_id=organization_id
        )
    )



from typing import Optional

from promptquality.get_metrics import get_metrics
from promptquality.helpers import (
    create_job,
    create_project,
    create_run,
    create_template,
    upload_dataset,
)
from promptquality.job_progress import job_progress
from promptquality.set_config import set_config
from promptquality.types.config import Config
from promptquality.types.run import PromptMetrics
from promptquality.types.settings import Settings
from promptquality.utils.dataset import DatasetType
from promptquality.utils.name import ts_run_name


def run(
    template: str,
    dataset: DatasetType,
    project_name: Optional[str] = None,
    run_name: Optional[str] = None,
    template_name: Optional[str] = None,
    settings: Optional[Settings] = None,
    wait: bool = True,
    config: Optional[Config] = None,
) -> Optional[PromptMetrics]:
    config = config or set_config()
    # Create project.
    project = create_project(project_name, config)
    # Create template.
    template_response = create_template(
        template,
        project.id,
        # Use project name as template name if not provided.
        template_name=template_name or project.name,
        config=config,
    )
    # Upload dataset.
    dataset_id = upload_dataset(
        dataset,
        project.id,
        template_response.selected_version_id,
        config,
    )
    # Run prompt.
    run_id = create_run(
        project.id,
        run_name=run_name
        or ts_run_name(
            template_response.name, template_response.selected_version.version
        ),
        config=config,
    )
    job_id = create_job(
        project.id,
        run_id,
        dataset_id,
        template_response.selected_version_id,
        settings,
        config,
    )
    if wait:
        job_progress(job_id, config)
    print(f"🔭 View your prompt run on the Galileo console at: {config.run_url}")
    metrics = get_metrics(
        project_id=project.id, run_id=run_id, job_id=job_id, config=config
    )
    return metrics



from itertools import product
from typing import List, Optional

from tqdm.auto import tqdm

from promptquality.helpers import create_project, get_estimated_cost
from promptquality.run import run
from promptquality.set_config import set_config
from promptquality.types.settings import Settings
from promptquality.utils.dataset import DatasetType


def create_settings_combinations(
    model_aliases: List[Optional[str]], temperatures: List[Optional[float]]
) -> List[Settings]:
    # Create all combinations of settings objects.
    model_aliases = model_aliases or [None]
    temperatures = temperatures or [None]
    return [
        Settings(model_alias=model_alias, temperature=temperature)
        for model_alias, temperature in product(model_aliases, temperatures)
    ]


def run_batch(
    templates: List[str],
    dataset: DatasetType,
    project_name: Optional[str] = None,
    model_aliases: Optional[List[str]] = None,
    temperatures: Optional[List[float]] = None,
    execute: bool = False,
    wait: bool = True,
) -> None:
    """
    Run a batch of prompts.
    """
    config = set_config()
    # Create project.
    project = create_project(project_name, config)

    # Estimate cost.
    estimated_costs = []
    aliases: List[Optional[str]] = list(model_aliases) if model_aliases else [None]
    all_temperatures: List[Optional[float]] = (
        list(temperatures) if temperatures else [None]
    )
    all_settings = create_settings_combinations(aliases, all_temperatures)
    all_combinations = list(product(templates, all_settings))
    for template, settings in all_combinations:
        estimated_costs.append(
            get_estimated_cost(
                dataset=dataset,
                template=template,
                settings=settings,
                project_id=project.id,
                config=config,
            )
        )
    print(
        f"Estimated total cost for {len(estimated_costs)} runs: "
        f"${sum(estimated_costs)}."
    )
    if not execute:
        print(
            "If you want to execute this run, invoke this function again with "
            "`execute=True`."
        )
    else:
        print(f"Running batch with {len(all_combinations)} runs...")
        for template, settings in tqdm(all_combinations):
            run(
                template=template,
                dataset=dataset,
                project_name=project.name,
                settings=settings,
                config=config,
                wait=wait,
            )



from typing import Optional

from pydantic import UUID4

from promptquality.set_config import set_config
from promptquality.types.config import Config
from promptquality.types.rows import GetRowsRequest, PromptRows


def get_rows(
    project_id: Optional[UUID4] = None,
    run_id: Optional[UUID4] = None,
    config: Optional[Config] = None,
) -> PromptRows:
    config = config or set_config()
    project_id = project_id or config.current_project_id
    if not project_id:
        raise ValueError("project_id must be provided")
    run_id = run_id or config.current_run_id
    if not run_id:
        raise ValueError("run_id must be provided")
    rows_json = config.api_client.get_rows(
        GetRowsRequest(project_id=project_id, run_id=run_id)
    )
    return PromptRows.model_validate(rows_json)



from typing import Dict, List, Optional

from pydantic import UUID4, BaseModel, Field


class GetRowsRequest(BaseModel):
    project_id: UUID4
    run_id: UUID4


class PromptRow(BaseModel):
    index: int
    prompt: Optional[str] = None
    response: Optional[str] = None
    target: Optional[str] = None
    inputs: Dict[str, Optional[str]] = Field(default_factory=dict)
    hallucination: Optional[float] = None
    bleu: Optional[float] = None
    rouge: Optional[float] = None
    cost: Optional[float] = None
    like_dislike: Optional[bool] = None


class PromptRows(BaseModel):
    rows: List[PromptRow]
